#
# Generic functions
#

# identity function
id(x) = x

# inequality relation
op(!=)(x, y) = ! (x == y)

# smaller or equal relation
op(<=)(x, y) = x < y || (x == y)

# greater relation
op(>)(x, y) = ! (x <= y)

# greater or equal relation
op(>=)(x, y) = ! (x < y)

#
# Boolean functions
#
# Boolean values are represented by the constructors False and True.
#

# logical conjunction
op(&&)(False, False) = False
op(&&)(False, True)  = False
op(&&)(True,  False) = False
op(&&)(True,  True)  = True

# logical disjunction
op(||)(False, False) = False
op(||)(False, True)  = True
op(||)(True,  False) = True
op(||)(True,  True)  = True

# logical negation
op(!)(False) = True
op(!)(True)  = False

# equality of booleans
op(==)(True,  True)  = True
op(==)(True,  False) = False
op(==)(False, False) = True
op(==)(False, True)  = False

# smaller relation on booleans
op(<)(True,  True)  = False
op(<)(True,  False) = False
op(<)(False, False) = False
op(<)(False, True)  = True

# convert boolean into string
show(False) = "False"
show(True)  = "True"

#
# Integer functions
#
# Integer values are represented as lists of boolean values that
# represent bits, with more significant bits towards the tail of
# list. The value 0 (zero) is represented as an empty list.
#
# Literal integers (0, 1, 42) are converted to lists by the parser.
#

# normalize integer
#
# This function ensures that there are no trailing zero bits at
# the end of a list. This is needed so that functions that make an
# integer smaller produce the right value for zero (an empty list)
# instead of arbitrary lists of zero bits which cannot be matched
# by a function guard.
#
normalize(x) =
  let dropFalse = \ x . case x of
                          []       = []
                          False:xs = self(xs)
                          _        = x
                        end
  in reverse(dropFalse(reverse(x)))

# addition of bit lists with carry bit
#
# This is a helper function for integer addition since it would be
# inconvenient to have to pass an initial value for the carry bit.
#
op(+)([], [], False)           = []
op(+)([], [], True)            = [True]
op(+)([], False:x, False)      = False : op(+)([], x, False)
op(+)([], False:x, True)       = True  : op(+)([], x, False)
op(+)([], True:x,  False)      = True  : op(+)([], x, False)
op(+)([], True:x,  True)       = False : op(+)([], x, True)
op(+)(False:x, [], False)      = False : op(+)([], x, False)
op(+)(False:x, [], True)       = True  : op(+)([], x, False)
op(+)(True:x,  [], False)      = True  : op(+)([], x, False)
op(+)(True:x,  [], True)       = False : op(+)([], x, True)
op(+)(False:x, False:y, False) = False : op(+)(x, y, False)
op(+)(False:x, False:y, True)  = True  : op(+)(x, y, False)
op(+)(True:x,  False:y, False) = True  : op(+)(x, y, False)
op(+)(True:x,  False:y, True)  = False : op(+)(x, y, True)
op(+)(False:x, True:y,  False) = True  : op(+)(x, y, False)
op(+)(False:x, True:y,  True)  = False : op(+)(x, y, True)
op(+)(True:x,  True:y,  False) = False : op(+)(x, y, True)
op(+)(True:x,  True:y,  True)  = True  : op(+)(x, y, True)

# addition of integers
op(+)(0, 0)             = 0
op(+)(0, x@(_:_))       = x
op(+)(x@(_:_), 0)       = x
op(+)(x@(_:_), y@(_:_)) = op(+)(x, y, False)

# subtraction of bit lists with carry bit
#
# This is a helper function for integer subtraction since it would be
# inconvenient to have to pass an initial value for the carry bit.
#
op(-)([], [], False)           = []
op(-)([], [], True)            = [True]
op(-)([], False:x, False)      = False : op(-)([], x, False)
op(-)([], False:x, True)       = True  : op(-)([], x, True)
op(-)([], True:x,  False)      = True  : op(-)([], x, False)
op(-)([], True:x,  True)       = False : op(-)([], x, False)
op(-)(False:x, [], False)      = False : op(-)([], x, False)
op(-)(False:x, [], True)       = True  : op(-)([], x, True)
op(-)(True:x,  [], False)      = True  : op(-)([], x, False)
op(-)(True:x,  [], True)       = False : op(-)([], x, False)
op(-)(False:x, False:y, False) = False : op(-)(x, y, False)
op(-)(False:x, False:y, True)  = True  : op(-)(x, y, True)
op(-)(True:x,  False:y, False) = True  : op(-)(x, y, False)
op(-)(True:x,  False:y, True)  = False : op(-)(x, y, False)
op(-)(False:x, True:y,  False) = True  : op(-)(x, y, True)
op(-)(False:x, True:y,  True)  = False : op(-)(x, y, True)
op(-)(True:x,  True:y,  False) = False : op(-)(x, y, False)
op(-)(True:x,  True:y,  True)  = True  : op(-)(x, y, True)

# subtracton of integers
#
# Negative results are in twos-complement format, meaning they have
# one bit more than any of the input values and that bit is always
# set. This means they are equivalent to a positive number of more
# bits than the input values.
#
op(-)(0, 0)             = []
op(-)(0, x@(_:_))       = normalize(op(-)(0, x, False))
op(-)(x@(_:_), 0)       = x
op(-)(x@(_:_), y@(_:_)) = normalize(op(-)(x, y, False))

# smaller relation of bit lists with carry bit
#
# This is a helper function for integer comparison since it would be
# inconvenient to have to pass an initial value for the carry bit.
#
op(<)([], [], True)            = True
op(<)([], [], False)           = False
op(<)([], False:x, True)       = op(<)([], x, True)
op(<)([], True:x,  True)       = True
op(<)([], False:x, False)      = op(<)([], x, False)
op(<)([], True:x,  False)      = True
op(<)(False:x, [], True)       = op(<)(x, [], True)
op(<)(True:x,  [], True)       = False
op(<)(False:x, [], False)      = op(<)(x, [], False)
op(<)(True:x,  [], False)      = False
op(<)(False:x, False:y, True)  = op(<)(x, y, True)
op(<)(False:x, True:y,  True)  = op(<)(x, y, True)
op(<)(True:x,  False:y, True)  = op(<)(x, y, False)
op(<)(True:x,  True:y,  True)  = op(<)(x, y, True)
op(<)(False:x, False:y, False) = op(<)(x, y, False)
op(<)(False:x, True:y,  False) = op(<)(x, y, True)
op(<)(True:x,  False:y, False) = op(<)(x, y, False)
op(<)(True:x,  True:y,  False) = op(<)(x, y, False)

# smaller relation on integers
op(<)(0, 0)                     = False
op(<)(0, x@(False:_))           = op(<)(0, x, False)
op(<)(0, x@(True:_))            = op(<)(0, x, False)
op(<)(x@(False:_), 0)           = op(<)(x, 0, False)
op(<)(x@(True:_), 0)            = op(<)(x, 0, False)
op(<)(x@(False:_), y@(False:_)) = op(<)(x, y, False)
op(<)(x@(False:_), y@(True:_))  = op(<)(x, y, False)
op(<)(x@(True:_),  y@(False:_)) = op(<)(x, y, False)
op(<)(x@(True:_),  y@(True:_))  = op(<)(x, y, False)

# equality of integers
op(==)(0, 0)             = True
op(==)(0, False:x)       = op(==)(0, x)
op(==)(0, True:_)        = False
op(==)(False:x, 0)       = op(==)(x, 0)
op(==)(True:_,  0)       = False
op(==)(False:x, False:y) = op(==)(x, y)
op(==)(False:_, True:_)  = False
op(==)(True:_,  False:_) = False
op(==)(True:x,  True:y)  = op(==)(x, y)

# shift integer bits to left
op(<<)(0, _) = 0
op(<<)(x, 0) = x
op(<<)(x, 1) = False : x
op(<<)(x, n) = False : (x << (n - 1))

# shift integer bits to right
op(>>)(_:x, 1) = x
op(>>)(x, n)   = drop(n, x)

# bitwise or of integers
op(|)(0, 0)               = 0
op(|)(0, x@(_:_))         = x
op(|)(x@(_:_), 0)         = x
op(|)(False:xs, False:ys) = False : (xs | ys)
op(|)(_:xs, _:ys)         = True  : (xs | ys)

# bitwise and of integers
op(&)(0, 0)             = 0
op(&)(0, _)             = 0
op(&)(_, 0)             = 0
op(&)(True:xs, True:ys) = True : normalize((xs & ys))
op(&)(_:xs, _:ys)       = normalize(False : (xs & ys))

# bitwise xor of integers
op(^)(0, 0)               = 0
op(^)(0, x@(_:_))         = x
op(^)(x@(_:_), 0)         = x
op(^)(False:xs, False:ys) = normalize(False : (xs ^ ys))
op(^)(True:xs, True:ys)   = normalize(False : (xs ^ ys))
op(^)(_:xs, _:ys)         = True : normalize(xs ^ ys)

# bitwise negation of integers
op(~)(0)        = 0
op(~)(False:xs) = True : normalize(~xs)
op(~)(True:xs)  = normalize(False : (~xs))

# multiplication of integers
op(*)(_, 0) = 0
op(*)(0, _) = 0
op(*)(x, False:y) = x * y << 1
op(*)(x, True:y)  = x + (x * y << 1)

# power of integers
#
# This function uses the exponentiation by squaring algorithm.
#
op(**)(_, 0)          = 1
op(**)(x, 1)          = x
op(**)(x, False:y)    = (x * x) ** y
op(**)(x, y@(True:_)) = x * ((x * x) ** ((y - 1) >> 1))

# division of integers
#
# This function works out the component powers of two of the
# quotient by repeated shifting and checking.
#
op(/)(x, y, q)      = if y * q == x then q else
                      if y * q <  x then op(/)(x, y, q << 1) else
                      q >> 1 + (x - (q >> 1 * y) / y)
op(/)(0, _)         = 0
op(/)(x, 1)         = x
op(/)(_:x, False:y) = x / y
op(/)(x, y)         = x < y ? 0 : op(/)(x, y, 1)

# remainder of integer division
op(%)(0, _:_) = 0
op(%)(x, y)   = x - ((x / y) * y)

# combination division and remainder
divmod(x,y) = let q = x / y, r = x - (q * y) in [q, r]

# factorial of an integer
fac(x) = case x of
           0 = 1
           n = n * fac(n - 1)
         end

# convert integer into string
show(0) = "0"
show(1) = "1"
show(2) = "2"
show(3) = "3"
show(4) = "4"
show(5) = "5"
show(6) = "6"
show(7) = "7"
show(8) = "8"
show(9) = "9"
show(x@(True:_))  = let [q, r] = divmod(x, 10) in show(q) ++ show(r)
show(x@(False:_)) = let [q, r] = divmod(x, 10) in show(q) ++ show(r)

#
# list functions
#
# List are represented by the constructor Nil for an empty list and
# by the constructor Cons(hd, tl) where hd is the first element of the
# list and tl is the rest of the list.
#
# Lists can be written as [ x, y, z ] which will be turned into the
# equivalent list Cons(x, Cons(y, Cons(z, Nil))). The Cons constructor
# can be abbreviated as infix : so that (x:xs) equals Cons(x,xs).
#

# test for empty list
null([])  = True
null(_:_) = False

# first element of list
head(x:_) = x

# all but the first element of list
tail(_:xs) = xs

# all but the last element of list
init([_])  = []
init(x:xs) = x : init(xs)

# last element of list
last([x])  = x
last(_:xs) = last(xs)

# length of a list
length([])   = 0
length(_:xs) = 1 + length(xs)

# concatenation of two lists
op(++)([], [])         = []
op(++)([], xs@(_:_))   = xs
op(++)(xs@(_:_), [])   = xs
op(++)(x:xs, ys@(_:_)) = x:(xs ++ ys)

# difference of two lists
op(--)([], [])         = []
op(--)([], _:_)        = []
op(--)(xs@(_:_), [])   = xs
op(--)(x:xs, ys@(_:_)) = elem(x, ys) ? (xs -- ys) : (x:(xs -- ys))

# equality of lists
op(==)([], [])     = True
op(==)([], _:_)    = False
op(==)(_:_, [])    = False
op(==)(x:xs, y:ys) = (x == y) && (xs == ys)

# smaller relation on lists
op(<)([], [])     = False
op(<)([], _:_)    = True
op(<)(_:_, [])    = False
op(<)(x:xs, y:ys) = x < y ? True : (x == y ? (xs < ys) : False)

# test for element in list
elem(_, [])   = False
elem(z, x:xs) = x == z ? True : elem(z, xs)

# get element from list at specific index
op(!!)(x:_,  0) = x
op(!!)(_:xs, n) = xs !! (n - 1)

# reverse elements in list
reverse(xs, [])   = xs
reverse(xs, y:ys) = reverse(y:xs, ys)
reverse(xs)       = reverse([], xs)

# drop number of elements from front of list
drop(0, [])       = []
drop(0, xs@(_:_)) = xs
drop(_, [])       = []
drop(n, x:xs)     = drop(n - 1, xs)

# take number of elements from front of list
take(0, [])   = []
take(0, _:_)  = []
take(_, [])   = []
take(n, x:xs) = x : take(n - 1, xs)

# intersperse element into list
intersperse(_, [])   = []
intersperse(_, [x])  = [x]
intersperse(e, x:xs) = x : e : intersperse(e, xs)

# create list of copies of a single value
replicate(0, _) = []
replicate(n, y) = y : replicate(n - 1, y)

# create list of sequence of integers
op(..)(x,y) = x >= y ? [x] : (x:(x + 1 .. y))

# turn list of lists into simple list
concat(xs@(_:_)) = foldr(op(++), [], xs)

# string containing only newline character
nl = [ 10 ]

# convert list into string
show(x@(_:_)) = "[" ++ concat(intersperse(", ", map(show, x))) ++ "]"

#
# higher order functions
#

# swap function arguments
flip(\ f) = \ x, y . f(y, x)

# function composition
op(*)(\ f, \ g) = \ x . f(g(x))

# apply function to all elements of a list
map(\ f, [])   = []
map(\ f, x:xs) = f(x):map(f, xs)

# filter elements from list
filter(\ f, [])   = []
filter(\ f, x:xs) = f(x) ? (x:filter(f, xs)) : filter(f, xs)

# fold list elements from the left
foldl(\ f, i, [])   = i
foldl(\ f, i, x:xs) = foldl(f, f(x, i), xs)

# fold list elements from the right
foldr(\ f, i, [])   = i
foldr(\ f, i, x:xs) = f(x, foldr(f, i, xs))

# drop elements from list while condition is met
dropWhile(\ f, [])   = []
dropWhile(\ f, x:xs) = f(x) ? dropWhile(f, xs) : (x:xs)

# take elements from list while condition is met
takeWhile(\ f, [])   = []
takeWhile(\ f, x:xs) = f(x) ? (x:takeWhile(f, xs)) : []

#
# Sorting functions
#

# split list in two halves by comparing with pivot element
pivot(p, []) = [[], []]
pivot(p, x:xs) =
  let [a, b] = pivot(p, xs)
  in x < p ? [x:a, b] : [a, x:b]

# sort list using the quicksort algorithm
qsort([]) = []
qsort(x:xs) =
  let [a, b] = pivot(x, xs)
  in qsort(a) ++ [x] ++ qsort(b)
